// ****************************************************************************
// copyright (c) 2000-2002 Horst Knorr <hk_classes@knoda.org>
// This file is part of the hk_classes library.
// This file may be distributed and/or modified under the terms of the
// GNU Library Public License version 2 as published by the Free Software
// Foundation and appearing in the file COPYING included in the
// packaging of this file.
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
// ****************************************************************************
//$Revision: 1.55 $
#include "hk_database.h"
#include "hk_data.h"
#include "hk_dbvisible.h"
#include "hk_presentation.h"
#include "hk_connection.h"
#include "hk_datasource.h"
#include "hk_actionquery.h"
#include "hk_form.h"
#include "hk_report.h"
#include "hk_dsmodevisible.h"

#include <sys/stat.h>
#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream.h>
#endif
#include <dirent.h>
#include <unistd.h>

newform_type *hk_database::p_newformfunction=NULL;
newreport_type *hk_database::p_newreportfunction=NULL;
newtable_type *hk_database::p_newtablefunction=NULL;
newquery_type *hk_database::p_newqueryfunction=NULL;


class hk_databaseprivate
{
public:
hk_databaseprivate(void)
			{
			  p_parent=NULL;
			}
        hk_string   p_dbname;
        hk_string p_sqldelimiter;
        list<hk_data*> p_hkdsourcelist;
        vector<hk_string> p_filelist;
        hk_string p_databasepath;
        list<hk_dbvisible*> p_visibles;
        list<hk_presentation*> p_presentations;
	void* p_parent;// will be used as widget parent in e.g.  ::new_formvisible;
};



hk_database::hk_database(hk_connection* c)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::hk_database");
#endif
    p_private=new hk_databaseprivate;
    p_connection=c;
}


hk_database::~hk_database()
{
#ifdef HK_DEBUG
    hkdebug("hk_database::~hk_database");
#endif
    clear_presentationlist();
    inform_datasources_before_closing();
    p_connection->db_remove(this);
    list<hk_data*>::iterator it;
    for (it=p_private->p_hkdsourcelist.begin();it!=p_private->p_hkdsourcelist.end();)
    {
        hk_data* dt=*it ;
        it++;
        delete dt;
    }
    p_private->p_hkdsourcelist.erase(p_private->p_hkdsourcelist.begin(),p_private->p_hkdsourcelist.end());
delete p_private;
}


bool hk_database::set_name(const hk_string& n)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::set_name");
#endif
hk_string nm=trim(n);
    if (nm==p_private->p_dbname)return true;
    hk_string oldname=p_private->p_dbname;
    p_private->p_dbname=nm;
    driver_specific_set_name();
    bool res=select_db();
    if (!res) p_private->p_dbname=oldname;
    return res;
}


hk_string hk_database::name(void)
{

    return p_private->p_dbname;
}


vector<hk_string>* hk_database::tablelist(void)
{
    return driver_specific_tablelist();
}


bool hk_database::table_exists(const hk_string& tablename)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::table_exists");
#endif
    bool res=false;
    vector<hk_string>* liste=tablelist();
    if (!liste) return false;
    vector<hk_string>::iterator it= liste->begin();
    while (it!=liste->end())
    {
        if ((*it)==tablename)return true;

        it++;
    }

    return res;
}


bool hk_database::query_exists(const hk_string& queryname)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::query_exists");
#endif
    bool res=false;
    vector<hk_string>* liste=querylist();
    if (!liste) return false;
    vector<hk_string>::iterator it= liste->begin();
    while (it!=liste->end())
    {
        if ((*it)==queryname)return true;

        it++;
    }

    return res;
}


hk_datasource* hk_database::new_table(const hk_string& name,hk_presentation* p)
{
if (!p_connection->database_exists(p_private->p_dbname)) return NULL;
    hk_datasource* tb=driver_specific_new_table(p);
    if (tb!=NULL)
    {
        p_private->p_hkdsourcelist.insert(p_private->p_hkdsourcelist.end(),tb);
        if (p_connection->server_needs(hk_connection::NEEDS_SQLDELIMITER))
            tb->set_sqldelimiter(p_connection->sqldelimiter());

    }
    else
    {
     show_warningmessage(hk_translate("Bug: hk_database::driver_specific_new_table returned empty table!"));
    }

    if (name.size()>0)tb->set_name(name);
    return tb;
}


hk_datasource* hk_database::new_resultquery(hk_presentation* p)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::new_resultquery");
#endif
if (!p_connection->database_exists(p_private->p_dbname)) return NULL;
    hk_datasource* tb=driver_specific_new_resultquery(p);
    if (tb!=NULL)
    {
        p_private->p_hkdsourcelist.insert(p_private->p_hkdsourcelist.end(),tb);
        if (p_connection->server_needs(hk_connection::NEEDS_SQLDELIMITER))
            tb->set_sqldelimiter(p_connection->sqldelimiter());
    }
    return tb;
}


hk_actionquery* hk_database::new_actionquery(void)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::new_actionquery");
#endif
if (!p_connection->database_exists(p_private->p_dbname)) return NULL;
    hk_actionquery* tb=driver_specific_new_actionquery();
    if (tb!=NULL) p_private->p_hkdsourcelist.insert(p_private->p_hkdsourcelist.end(),tb);
    return tb;
}


void hk_database::ds_remove(hk_data* ds)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::ds_remove");
#endif

    p_private->p_hkdsourcelist.remove(ds);

}


bool hk_database::select_db(void)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::select_db");
#endif
    inform_datasources_before_closing();
    if (p_connection->is_connected())
    {
        bool erg= driver_specific_select_db();
        if (erg)
        {
            p_private->p_databasepath=p_connection->databasepath();
            p_private->p_databasepath+="/";
            p_private->p_databasepath+=name();
            mkdir (p_private->p_databasepath.c_str(),S_IRUSR|S_IWUSR|S_IXUSR);
            hk_string b=p_private->p_databasepath+"/output";
            mkdir (b.c_str(),S_IRUSR|S_IWUSR|S_IXUSR);

        }
        else
            show_warningmessage(hk_translate("No such Database!"));
        return erg;
    }
    else
    {
        show_warningmessage(hk_translate("Not connected to server!"));
        return false;
    }

}


hk_string hk_database::database_path(void)
{

    return p_private->p_databasepath;
}


void hk_database::connection_connected(void)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::connection_connected");
#endif
    select_db();

}


void hk_database::connection_disconnected(void)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::connection_disconnected");
#endif
    list<hk_data*>::iterator it ;
    it=p_private->p_hkdsourcelist.begin();
    mark_datasources_as_not_handled();
    while (it!=p_private->p_hkdsourcelist.end())
    {
        if (!(*it)->p_already_handled)
        {
            (*it)->before_source_vanishes();
            (*it)->before_connection_disconnects();
            (*it)->p_already_handled=true;
            it=p_private->p_hkdsourcelist.begin();
        }
        else
            it++;
    }
}


bool hk_database::delete_table(const hk_string& table,enum_interaction x)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::delete_table");
#endif
    if (!p_connection->is_connected()) return false;
    hk_actionquery* q=new_actionquery();
    hk_string sql="DROP TABLE ";
    hk_string del;
    hk_datasource* d=new_table();
    if (d)
      {
       del=d->p_identifierdelimiter;
       delete d;
      }
    sql+=del+table+del+p_private->p_sqldelimiter;
    bool result=false;
    hk_string warning=hk_translate("Delete the table \"%TBLNAME%\"?");
    warning=replace_all("%TBLNAME%",warning,table);
    if(x==noninteractive||show_yesnodialog(warning,true))
    {

        if (q!=NULL)
        {
            hk_datasource* d=new_table();
            d->set_name(table);
            d->driver_specific_before_drop_table();
            delete d;
            q->set_sql(sql.c_str(),sql.size());
            result=q->execute();
            delete  q;
        }
    } else return false;
    if (result)
    {
        list<hk_data*>::iterator it ;
        it=p_private->p_hkdsourcelist.begin();
        while (it!=p_private->p_hkdsourcelist.end())
        {
            hk_data* vi=*it;
            hk_string t=table;
            if (vi->type()==hk_data::ds_table && vi->name()==t)vi->before_source_vanishes();
            it++;
        }

    }
    else
    {
        hk_string reason=hk_translate("Table could not be deleted!")+"\n"+hk_translate("Servermessage: ")+connection()->last_servermessage();
        show_warningmessage(reason);
    }
    return result;

}


void hk_database::inform_datasources_before_closing(void)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::inform_datasources_before_closing");
#endif
    mark_datasources_as_not_handled();
    list<hk_data*>::iterator it ;
    it=p_private->p_hkdsourcelist.begin();
    while (it!=p_private->p_hkdsourcelist.end())
    {
        hk_data* vi=(*it);
        ++it;
        if (!vi->p_already_handled)
        {
            vi->before_source_vanishes();
            vi->p_already_handled=true;
            it=p_private->p_hkdsourcelist.begin();
        }

    }

}


void hk_database::save(const hk_string& statement,const hk_string& name,filetype type)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::save");
#endif
    hk_string n=p_private->p_databasepath+"/";
    n+=name+fileendings(type);
//   cout <<n<<endl;
    FILE* out=fopen(n.c_str(),"w");

    fwrite(statement.c_str(),statement.size(),1,out);
    fclose(out);
//   if (r<(signed)res->length)return false;
//return true;

}


ofstream* hk_database::savestream(const hk_string&name,filetype type,bool ask_before_overwrite)
{
    return savestream(name,type,ask_before_overwrite,true,false);
}


ofstream* hk_database::savestream(const hk_string&name,filetype type,bool ask_before_overwrite,bool with_header,bool ask_for_new_name)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::savestream");
#endif
    hk_string filename=p_private->p_databasepath+"/";
    filename+=name+fileendings(type);
    hk_string dialog;
    if (type==ft_query)dialog=hk_translate("Query already exists. Overwrite it?");
    else if (type==ft_form)dialog=hk_translate("Form already exists. Overwrite it?");
    else if (type==ft_report)dialog=hk_translate("Report already exists. Overwrite it?");
    else dialog=hk_translate("File already exists. Overwrite it?");
    if (ask_before_overwrite)
    {
        ifstream ifs(filename.c_str(),ios::in);
        if (ifs)
        {
//file already exists, overwrite it?
            if(!show_yesnodialog(dialog,true))
            {
// no, so don't overwrite it
                if (ask_for_new_name)
                {
                    if (type==ft_query)dialog=hk_translate("Enter new query name:");
                    else if (type==ft_form)dialog=hk_translate("Enter new form name:");
                    else if (type==ft_report)dialog=hk_translate("Enter new report name:");
                    else dialog=hk_translate("Enter new name:");
                    hk_string res=show_stringvaluedialog(dialog);
                    if (res.size()==0) return NULL;
                    return savestream(res,type,ask_before_overwrite,with_header,ask_for_new_name);

                }
                else
                    return NULL;
            }
        }
    }

    ofstream* stream= new ofstream(filename.c_str());
    if (stream!=NULL)
    {
        inform_datasources_filelist_changes(type);
        if (with_header) (*stream) <<"<?xml version=\"1.0\" encoding=\"UTF-8\" ?>"<<endl;
    }
    return stream;

}


hk_string hk_database::load(const hk_string& name,filetype type)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::load");
#endif
    if (name.size()==0) return "";
    hk_string n=p_private->p_databasepath+"/";
    n+=name+fileendings(type);
    hk_string result;
    ifstream ifs(n.c_str(),ios::in);
    if (ifs)
    {
        char c;
        while (ifs.get(c))
            result+=c;
    }
    else cerr <<"no file: "<<n<<endl;
    return result;
}


hk_string hk_database::fileendings(filetype e)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::fileendings");
#endif
    switch (e)
    {

        case ft_query: return ".hk_query";
        case ft_table: return ".hk_table";
        case ft_form: return ".hk_form";
        case ft_report: return ".hk_report";
        default : return ".hk_unknown";
    }
    return ".hk_unknown";
}


vector<hk_string>* hk_database::querylist(void)
{
    return filelist(ft_query);
}


vector<hk_string>* hk_database::formlist(void)
{
    return filelist(ft_form);
}


vector<hk_string>* hk_database::reportlist(void)
{
    return filelist(ft_report);
}


/*     DIR *dp;
     struct dirent *entry;
     hk_string datei;

 p_querylist.erase(p_querylist.begin(),p_querylist.end());
   if ((dp=opendir(p_private->p_databasepath.c_str()))!=NULL)
   {

     while ((entry = readdir(dp))!=NULL)
     {
       datei=entry->d_name;
size_t p =datei.find(fileendings(ft_query));
if (p<datei.size())
{
datei.replace(p,datei.size()-p,"");
p_querylist.insert(p_querylist.end(),datei);
}
}
}

return &p_querylist;
*/
//}

vector<hk_string>* hk_database::filelist(filetype type)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::filelist");
#endif
    DIR *dp;
    struct dirent *entry;
    hk_string datei;

    p_private->p_filelist.erase(p_private->p_filelist.begin(),p_private->p_filelist.end());
    if ((dp=opendir(p_private->p_databasepath.c_str()))!=NULL)
    {

        while ((entry = readdir(dp))!=NULL)
        {
            datei=entry->d_name;
            size_t p =datei.find(fileendings(type));
            if (p<datei.size())
            {
                datei.replace(p,datei.size()-p,"");
                p_private->p_filelist.insert(p_private->p_filelist.end(),datei);
            }
        }
        closedir(dp);

    }
    sort(p_private->p_filelist.begin(),p_private->p_filelist.end());
    return &p_private->p_filelist;
}


bool  hk_database::delete_file(const hk_string& name,filetype type,enum_interaction x)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::delete_file");
#endif
    hk_string filename=p_private->p_databasepath+"/";
    filename+=name+fileendings(type);
    hk_string message;
    if (type==ft_query)
        message=hk_translate("Delete the query:%FNAME%?");
    else
    if (type==ft_report)
        message=hk_translate("Delete the report:%FNAME%?");
    else
    if (type==ft_form)
        message=hk_translate("Delete the form:%FNAME%?");
    else
        message=hk_translate("Delete the file:%FNAME%");
    message=replace_all("%FNAME%",message,name);
    if(x==interactive&&!show_yesnodialog(message,true))
        return false;
    int r=unlink(filename.c_str());
    if (r==0)
    {
        inform_datasources_filelist_changes(type);
        return true;
    }
    return false;
}


void hk_database::inform_datasources_filelist_changes(filetype type)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::inform_datasources_filelist_changes");
#endif
    list<hk_data*>::iterator it ;
    it=p_private->p_hkdsourcelist.begin();
    while (it!=p_private->p_hkdsourcelist.end())
    {
        hk_data* vi=*it;
        it++;
        vi->filelist_changes(type);
    }

}


hk_datasource*  hk_database::load_datasource(const hk_string& name,bool query,hk_presentation* p)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::load_datasource");
#endif
    if (query)
    {
        hk_datasource* d=NULL;
        hk_string r=load(name,ft_query);
        if (r.size()>0)
        {
            d=new_resultquery(p);
        }
        if (d!=NULL) d->loaddata(u2l(r));
        return d;
    }
    else
        return new_table(name,p);

}


hk_datasource* hk_database::driver_specific_new_table(hk_presentation* p)
{
    return NULL;
}


hk_datasource* hk_database::driver_specific_new_resultquery(hk_presentation* p)
{
    return NULL;
}


void hk_database::dbvisible_add(hk_dbvisible* v)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::visible_add");
#endif
    p_private->p_visibles.insert(p_private->p_visibles.end(),v);
}


void hk_database::dbvisible_remove(hk_dbvisible* v)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::visible_remove");
#endif
    p_private->p_visibles.remove(v);
}


void hk_database::clear_visiblelist(void)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::clear_visiblelist");
#endif
    mark_visible_objects_as_not_handled();
    list<hk_dbvisible*>::iterator it=p_private->p_visibles.begin();
    while ( it!=p_private->p_visibles.end() )
    {
        hk_dbvisible* dv=*it;
        ++it;
        if (!dv->p_already_handled)
        {
            dv->p_already_handled=true;
            dv->database_delete();
            p_private->p_visibles.remove(dv);
            it=p_private->p_visibles.begin();

        }
    }

}


void hk_database::before_source_vanishes(void)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::before_source_vanishes");
#endif
    list<hk_dbvisible*>::iterator it =p_private->p_visibles.begin();
    while(it!=p_private->p_visibles.end())
    {
        hk_dbvisible* vi=(*it);
        vi->before_source_vanishes();
        it=find(p_private->p_visibles.begin(),p_private->p_visibles.end(),vi);
        if (it==p_private->p_visibles.end()) it=p_private->p_visibles.begin();
        else it++;
    }
#ifdef HK_DEBUG
    hkdebug("ENDE database::before_source_vanishes");
#endif
}


void hk_database::presentation_add(hk_presentation* p)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::presentation_add");
#endif
    p_private->p_presentations.insert(p_private->p_presentations.end(),p);
}


void hk_database::presentation_remove(hk_presentation* p)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::presentation_remove");
#endif
    p_private->p_presentations.remove(p);
}


void hk_database::clear_presentationlist(void)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::clear_presentationlist");
#endif
    list<hk_presentation*>::iterator it;
    it=p_private->p_presentations.begin();
    while ( it!=p_private->p_presentations.end() )
    {
        hk_presentation* dv=*it;
#ifdef HK_DEBUG
        hkdebug("dv->classname: ");
        hkdebug(dv->hkclassname());
#endif
        ++it;
        dv->database_delete();
        p_private->p_presentations.remove(dv);
    }

}


hk_presentation* hk_database::existing_presentation(const hk_string& name, hk_presentation::enum_presentationtype t)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::existing_presentation");
#endif
    list<hk_presentation*>::iterator it;
    it=p_private->p_presentations.begin();
    while ( it!=p_private->p_presentations.end() )
    {
        hk_presentation* dv=*it;
        ++it;
        if (dv->name()==name &&dv->presentationtype()== t)
        {
            return dv;
        }
    }
    return NULL;
}


bool hk_database::copy_table(hk_datasource* fromdatasource,bool schema_and_data,bool replacetable, bool ask,progress_dialogtype* progressdialog)
{
#ifdef HK_DEBUG
    hkdebug("hk_database::copy_table");
#endif
    if (!fromdatasource)return false;
    if (fromdatasource->columns()->size()==0)return false;
    bool tblexists=table_exists(fromdatasource->name());
    if (tblexists&&!ask&&!replacetable)return false;
    hk_datasource* newds=new_table();
    hk_string structure,buffer;
#ifdef HAVE_SSTREAM
    stringstream stream;
#else
    strstream stream;
#endif
    fromdatasource->save_datasourcedefinition(stream);
    while (stream >>buffer) structure+=buffer;
    bool result=newds->load_datasourcedefinition(u2l(structure),replacetable);
    bool cancel=false;
    if (result&&schema_and_data)
    {
        newds->set_accessmode(hk_datasource::batchwrite);
        newds->enable();
        if (fromdatasource->is_enabled())fromdatasource->disable();
        hk_datasource::enum_accessmode access=fromdatasource->accessmode();
        fromdatasource->set_accessmode(hk_datasource::batchread);
        fromdatasource->enable();
        unsigned long r=0;
	if (!newds->columns())
	  {
	   if (ask) show_warningmessage(hk_translate("hk_database::copy_table: bug newds has no columns!"));
	   return false;
	   }
        list<hk_column*>::iterator colit =newds->columns()->begin();
        while (colit!=newds->columns()->end())
        {
            (*colit)->set_allow_autoincwrite(true);
            colit++;
        }

        while (r<fromdatasource->max_rows()&&!cancel)
        {
            newds->setmode_insertrow();
            list<hk_column*>::iterator it=fromdatasource->columns()->begin();
            while (it!=fromdatasource->columns()->end())
            {
                hk_column* c=newds->column_by_name((*it)->name(),fromdatasource->columnname_occurance(*it));
                if (c&&!(*it)->is_nullvalue())
                {

                    switch(c->columntype())
                    {
                        case hk_column::binarycolumn: c->set_asbinary((*it)->asbinary());break;
                        case hk_column::boolcolumn: c->set_asbool((*it)->asbool());break;
                        default: if ((*it)->asstring().size()>0)
                        {

                            c->set_asstring((*it)->asstring());
                        }
                    }
                }

                it++;
            }

            if (progressdialog) cancel=progressdialog(r,fromdatasource->max_rows(),replace_all("%TABLE%",hk_translate("Copying table %TABLE%"),name()));
            newds->store_changed_data();
            fromdatasource->goto_next();
            r++;
        }

        newds->disable();
        fromdatasource->disable();
        fromdatasource->set_accessmode(access);
        delete newds;
    }

    if (result)inform_datasources_filelist_changes(ft_table);
    if (cancel) return false;
    return result;
}


void    hk_database::mark_datasources_as_not_handled(void)
{
    list<hk_data*>::iterator it=p_private->p_hkdsourcelist.begin();
    while (it!=p_private->p_hkdsourcelist.end())
    {
        (*it)->p_already_handled=false;
        ++it;
    }

}


void    hk_database::mark_visible_objects_as_not_handled(void)
{
    list<hk_dbvisible*>::iterator it=p_private->p_visibles.begin();
    while (it!=p_private->p_visibles.end())
    {
        (*it)->p_already_handled=false;
        ++it;
    }

}

hk_form* hk_database::new_formvisible(void)
{
   hk_form* result=NULL;
    if (p_newformfunction)
    	result=p_newformfunction(this,p_private->p_parent);
    else
    {
      result=new hk_form();
      if (result) result->set_database(this);
    }
   return result;
}



hk_report* hk_database::new_reportvisible(void)
{
   hk_report* result=NULL;
    if (p_newreportfunction)
    	result=p_newreportfunction(this,p_private->p_parent);
    else
    {
     result=new hk_report();
     if (result) result->set_database(this);
    }
   return result;
}

void	hk_database::set_new_formvisiblefunction(newform_type* f)
{
p_newformfunction=f;
}

void	hk_database::set_new_reportvisiblefunction(newreport_type* f)
{
p_newreportfunction=f;

}


void	hk_database::set_new_tablevisiblefunction(newtable_type* f)
{
p_newtablefunction=f;

}


void	hk_database::set_new_queryvisiblefunction(newquery_type* f)
{
p_newqueryfunction=f;

}


hk_dstable* hk_database::new_tablevisible(void)
{
    if (p_newtablefunction)
    	 return p_newtablefunction(p_private->p_parent);
    else return new hk_dstable();
}

hk_dsquery* hk_database::new_queryvisible(void)
{
    if (p_newqueryfunction)
    	 return p_newqueryfunction(p_private->p_parent);
    else return new hk_dsquery();
}

void hk_database::set_usewidgetparent(void* parent)
{
 p_private->p_parent=parent;
}

bool hk_database::rename_table(const hk_string& originalname, const hk_string& new_name,enum_interaction x)
{
   if ( driver_specific_rename_table(originalname,new_name))
   {
     list<hk_data*>::iterator it=p_private->p_hkdsourcelist.begin();
     while (it!=p_private->p_hkdsourcelist.end())
     {
       if ((*it)->type()==hk_data::ds_table && (*it)->name()==originalname)
           {
		bool enabled=((hk_datasource*)(*it))->is_enabled();
		if (enabled) ((hk_datasource*)(*it))->disable();
		(*it)->set_name(new_name);
		if (enabled) ((hk_datasource*)(*it))->enable();
	   }
       ++it;
       inform_datasources_filelist_changes(ft_table);
     }


   return true;
   }
   else return false;
}


bool hk_database::rename_file(const hk_string& originalname, const hk_string& new_name,filetype ftype,enum_interaction x)
{
    if (ftype==ft_table) return rename_table(originalname,new_name,x);
    hk_string n=p_private->p_databasepath+"/"+originalname+fileendings(ftype);

    ifstream ifs(n.c_str(),ios::in);
    ofstream* ofs = NULL;
    if (ifs)ofs=savestream(new_name,ftype,true,false,true);

    if (ifs && ofs &&(*ofs))
    {                                             //copy file
        char c;
        while (ifs.get(c))
            ofs->put(c);
    }
    else
    {
        if (!ifs && x==interactive) show_warningmessage("no file: "+n);
        if (ofs) delete ofs;
        return false;
    }
    ofs->close();
    ifs.close();
    delete ofs;
    delete_file(originalname,ftype,noninteractive);
    return true;



}


